package org.iotivity.bridge.homeserver;

import android.app.AlarmManager;
import android.app.IntentService;
import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.os.Binder;
import android.os.Build;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.os.Process;
import android.os.RemoteCallbackList;
import android.os.RemoteException;
import android.support.annotation.Nullable;
import android.support.v4.app.NotificationCompat;
import android.util.Log;
import android.widget.Toast;


import org.iotivity.OCBridge;
import org.iotivity.OCCoreRes;
import org.iotivity.OCCred;
import org.iotivity.OCCredType;
import org.iotivity.OCCredUtil;
import org.iotivity.OCCreds;
import org.iotivity.OCUuid;
import org.iotivity.OCUuidUtil;
import org.iotivity.bridge.homeserver.handler.client.DeviceDiscoveryHandler;
import org.iotivity.bridge.homeserver.model.Bridge;
import org.iotivity.oc.OcRemoteDevice;
import org.iotivity.oc.OcUtils;

import java.util.Calendar;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

public class BridgeService extends IntentService {

    public static final String TAG = BridgeService.class.getName();

    public static Intent serviceIntent = null ;
    private final IBinder binder = new LocalServiceBinder();
    private BridgeServer bridgeServer = null ;

    private ScheduledExecutorService executorService = Executors.newSingleThreadScheduledExecutor();

    // IRemoteService MSG TYPEs
    public static final int MSG_REPORT = 1 ;
    public static final int MSG_STRING = 2;
    public static final int MSG_LINE = 3;
    public static final int MSG_DEVICE_DISCOVERED = 4 ;

    final RemoteCallbackList<IRemoteServiceCallback> mCallbacks
            = new RemoteCallbackList<IRemoteServiceCallback>();

    public BridgeService() {
        super(BridgeService.class.getName());
    }

    private BridgeService getService() {
        return this ;
    }

    @Override
    public void onCreate() {
        super.onCreate();

        Log.d(TAG, "onCreate()");
        // start first time only
        bridgeServer = new BridgeServer(this);

        scheduledTaskInit();
    }

    /*
     * BridgeService can be binded with Local Activity and Remote Application
     * through RemoteInterface and ServiceBinder
     */
    @Override
    public IBinder onBind(Intent intent) {

        Log.d(TAG, "onBind() action = " + intent.getAction());
        if (IRemoteService.class.getName().equals(intent.getAction())) {
            return mBinder;
        }
        return binder;
    }

    /*
     * This is to bind Local Activity like Local Service
     */
    public class LocalServiceBinder extends Binder {
        public BridgeService getService() {
            return BridgeService.this ;
        }

        public int getPid() {
            return Process.myPid();
        }
    }

    /**
     * The IRemoteInterface is defined through IDL
     */
    private final IRemoteService.Stub mBinder = new IRemoteService.Stub() {
        public void registerCallback(IRemoteServiceCallback cb) {
            if (cb != null) mCallbacks.register(cb);

            mHandler.sendMessage(mHandler.obtainMessage(MSG_STRING, "Remote Client Callback was registered."));
            mHandler.sendEmptyMessage(MSG_LINE);

        }
        public void unregisterCallback(IRemoteServiceCallback cb) {
            if (cb != null) mCallbacks.unregister(cb);
        }

        @Override
        public int getPid() throws RemoteException {
            return Process.myPid();
        }

    };

    private final Handler mHandler = new Handler() {
        @Override public void handleMessage(Message msg) {
            switch (msg.what) {

                case MSG_LINE: {

                    // Broadcast to all clients the new value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).lineMessageNotified();
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                            Log.e(TAG, "RemoteException :", e);
                        }
                    }
                    mCallbacks.finishBroadcast();

                } break;
                case MSG_STRING : {
                    // Broadcast to all clients the new value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).messageNotified((String)msg.obj);
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                            Log.e(TAG, "RemoteException :", e);
                        }
                    }
                    mCallbacks.finishBroadcast();
                } break ;
                case MSG_DEVICE_DISCOVERED: {
                    // Broadcast to all clients the new value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).addDeviceNotified();
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                            Log.e(TAG, "RemoteException :", e);
                        }
                    }
                    mCallbacks.finishBroadcast();
                } break ;
                default:
                    super.handleMessage(msg);
            }
        }
    };

    public void msg(String msg) {
        mHandler.sendMessage(mHandler.obtainMessage(MSG_STRING, msg));
    }

    public void printLine() {
        mHandler.sendEmptyMessage(MSG_LINE);
    }

    @Override
    protected void onHandleIntent(@Nullable Intent intent) {

        Log.d(TAG, "onHandleIntent() intent=" + intent.getAction());
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {

        Log.d(TAG, "onStartCommand() intent=" + intent.getAction());
        serviceIntent = intent ;

        initializeNotification();

        return START_STICKY;
    }

    @Override
    public void onDestroy() {
        super.onDestroy();

        Log.d(TAG, "onDestroy()");

        Log.d(TAG, "Calling Bridge Shutdown.");
        bridgeServer.shutdown();

        // Unregister all callbacks.
        mCallbacks.kill();

        // Remove the next pending message to increment the counter, stopping
        // the increment loop.
        //mHandler.removeMessages(MSG_REPORT);

        // Bridge Service and Application Process Kill
        // android.os.Process.killProcess(Process.myPid());
        // or below action
        Log.d(TAG, "System.exit() will be called");
        System.runFinalizersOnExit(true);
        System.exit(0);
    }

    @Override
    public void onTaskRemoved(Intent rootIntent) {
        super.onTaskRemoved(rootIntent);

        Log.d(TAG, "onTaskRemoved()");

        this.restartServiceByAlarmManager();
    }

    @Override
    public void onStart(@Nullable Intent intent, int startId) {
        super.onStart(intent, startId);
        Log.d(TAG, "Bridge Service onStart()");
    }

    @Override
    public boolean onUnbind(Intent intent) {
        Log.d(TAG, "Bridge Service onUnbind()");

        return super.onUnbind(intent);
    }

    private void initializeNotification() {

        NotificationCompat.Builder builder = new NotificationCompat.Builder(this, "1");
        builder.setSmallIcon(R.mipmap.ic_launcher);
        NotificationCompat.BigTextStyle style = new NotificationCompat.BigTextStyle();
        style.bigText("Press this to see settings.");
        style.setBigContentTitle(null);
        style.setSummaryText("Bridge Service is Running.");
        builder.setContentText(null);
        builder.setContentTitle(null);
        builder.setOngoing(true);
        builder.setStyle(style);
        builder.setWhen(0);
        builder.setShowWhen(false);

        Intent notificationIntent = new Intent(this, BridgeServerActivity.class);
        PendingIntent pendingIntent = PendingIntent.getActivity(this, 0, notificationIntent, 0);
        builder.setContentIntent(pendingIntent);

        NotificationManager manager = (NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE);
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            manager.createNotificationChannel(new NotificationChannel("1", "bridge_service", NotificationManager.IMPORTANCE_NONE));
        }
        Notification notification = builder.build();
        startForeground(1, notification);
    }


    /* Becuase stopSrvice() on Destroy() will kill all application process of service, activity clearly
     * Do not use this function for undead foreground service. It's not meaningful */
    private void restartServiceByAlarmManager() {

        final Calendar calendar = Calendar.getInstance();
        calendar.setTimeInMillis(System.currentTimeMillis());
        calendar.add(Calendar.SECOND, 3);
        Intent intent = new Intent(this, AlarmReceiver.class);
        PendingIntent sender = PendingIntent.getBroadcast(this, 0,intent,0);
        AlarmManager alarmManager = (AlarmManager) getSystemService(Context.ALARM_SERVICE);
        alarmManager.set(AlarmManager.RTC_WAKEUP, calendar.getTimeInMillis(), sender);
    }

    public BridgeServer getBridgeServer() {
        return bridgeServer ;
    }

    public void scheduledTaskInit () {

        // How to use ExecuterService like followings
        // executorService.schedule(Classname::someTask, delayInSeconds, TimeUnit.SECONDS);                 // once after a delay
        // executorService.scheduleAtFixedRate(Classname::someTask, 0, delayInSeconds, TimeUnit.SECONDS);   // at fixed time interval
        executorService.scheduleAtFixedRate(new Runnable() {
            @Override
            public void run() {
                if(bridgeServer!=null) bridgeServer.loadPairwisedDeviceUUIDs();

                clearPairwisedDevices();
                OCUuid bridge_di = OCCoreRes.getDeviceId(Bridge.BRIDGE_INDEX);
                if (!OcUtils.discoverAllDevices(new DeviceDiscoveryHandler(getService(), bridge_di))) {
                    final String msg = "Failed to discover devices";
                    Log.d(TAG, msg);
                }
                Log.d(TAG, "DeviceDiscovery ScheduledTask per 60sec was called.");
            }
        }, 10, 60, TimeUnit.SECONDS); // 60sec interval after 10sec delay

        /*executorService.scheduleAtFixedRate(new Runnable() {
            @Override
            public void run() {
                // TODO
                bridgeServer.loadPairwisedDeviceUUIDs();
            }
        }, 30, 60, TimeUnit.SECONDS); // 30sec interval after 30sec delay
*/
        /*executorService.schedule(new Runnable() {
            @Override
            public void run() {

            }
        }, 10, TimeUnit.SECONDS);*/
    }

    public void clearPairwisedDevices() {
        if(bridgeServer!=null) {
            bridgeServer.clearPairwisedDevices();
        }
    }

    public void addPairwisedDevice(OcRemoteDevice device) {
        if(bridgeServer!=null) {
            bridgeServer.addOrReplacePairwisedDevice(device);
            mHandler.sendEmptyMessage(MSG_DEVICE_DISCOVERED);
        }
    }

    public boolean hasPairwisedDeviceUUIDs(OCUuid uuid) {
        if(bridgeServer!=null) {
            return bridgeServer.hasPairwisedDeviceUUIDs(uuid);
        }
        return false ;
    }

}
